from argparse import Namespace
from typing import Iterable, List, Set
from urllib.parse import urldefrag, urljoin, urlparse

import requests
from bs4 import BeautifulSoup as BS

from wbsv.url_filters import CombinedFilter, OwnDomainFilter, SchemaFilter, UrlFilter


class MissingURLSchemaWarning(UserWarning):
    pass


class Crawler:
    @staticmethod
    def from_parser_args(args: Namespace) -> "Crawler":
        return Crawler.from_args(args.url, args.own, args.only_target, args.level)

    @staticmethod
    def from_args(
        urls: Iterable[str], own: bool, only_target: bool, level: int
    ) -> "Crawler":
        schema_filter = SchemaFilter()
        normalized_urls = Crawler._normalize_urls(urls, schema_filter)
        domain_filters = (
            [schema_filter, OwnDomainFilter.from_string_urls(set(normalized_urls))]
            if own
            else [schema_filter]
        )
        return Crawler(
            urls=normalized_urls,
            domain_filter=CombinedFilter(domain_filters),
            only_target=only_target,
            level=level,
        )

    def __init__(
        self, urls: List[str], domain_filter: UrlFilter, only_target: bool, level: int
    ):
        """Init."""
        self.urls = urls
        self.domain_filter = domain_filter
        self.only_target = only_target
        self.level = level
        self.queue: List[Set[str]] = []
        self.UA: str = (
            "Mozilla/5.0 (Windows NT 5.1; rv:40.0) " "Gecko/20100101 Firefox/40.0"
        )

    def run_crawl(self) -> List[Set[str]]:
        """Execute crawler."""
        if self.only_target:
            return [set(self.urls)]
        for now_level in range(self.level):
            self._crawl(now_level)

        return self.queue

    def _crawl(self, now_level: int) -> None:
        """Helper for crawling."""
        collecting_links = set()
        collected_links: Set[str] = set()
        collected_links = collected_links.union(*self.queue)
        if now_level == 0:
            self.queue.append(set(self.urls))
        for url in self.queue[-1]:
            source = requests.get(url, headers={"User-Agent": self.UA}).content
            data = BS(source, features="lxml")
            extracted_urls = [
                urljoin(url, _.get("href")) for _ in set(data.find_all("a"))
            ]
            collecting_links |= set(
                Crawler._normalize_urls(extracted_urls, url_filter=self.domain_filter)
            )
        self.queue.append(collecting_links - collected_links)

    @staticmethod
    def _normalize_urls(urls: Iterable[str], url_filter: UrlFilter) -> List[str]:
        """Normalize parsed_url."""
        valid_urls = []
        for url in urls:
            parsed_url = urlparse(url)
            if url_filter.test_url(parsed_url):
                valid_urls.append(urldefrag(parsed_url.geturl()).url)
        return valid_urls



from typing import Iterable, Set, TypedDict
from urllib.parse import urlparse

from wbsv import crawler, main

# UrlFilter unit tests
from wbsv.url_filters import OwnDomainFilter, SchemaFilter


class ArgDict(TypedDict):
    urls: Iterable[str]
    own: bool
    only_target: bool
    level: int


def test_schema_filter_should_reject_invalid_schema() -> None:
    schema_filter = SchemaFilter()
    invalid_schema_url = urlparse("data://test.txt")

    assert schema_filter.test_url(invalid_schema_url) is False


def test_schema_filter_should_accept_valid_schema() -> None:
    schema_filter = SchemaFilter()
    valid_schema_url = urlparse("https://example.com")

    assert schema_filter.test_url(valid_schema_url) is True


def test_own_domain_filter_should_reject_url_with_invalid_domain() -> None:
    own_domain_filter = OwnDomainFilter.from_string_urls(["https://example.com"])
    invalid_domain_url = urlparse("https://www.iana.org")

    assert own_domain_filter.test_url(invalid_domain_url) is False


def test_own_domain_filter_should_accept_url_with_valid_domain() -> None:
    own_domain_filter = OwnDomainFilter.from_string_urls(["https://example.com"])
    invalid_domain_url = urlparse("https://example.com")

    assert own_domain_filter.test_url(invalid_domain_url) is True


def test_api_without_parser() -> None:
    args: ArgDict = {
        "urls": ["https://example.com"],
        "own": False,
        "only_target": False,
        "level": 1,
    }
    c = crawler.Crawler.from_args(**args)
    urls = {"https://example.com", "https://www.iana.org/domains/example"}
    t: Set[str] = set()
    t = t.union(*c.run_crawl())
    if urls != t:
        raise AssertionError


def test1() -> None:
    args = main.parse_args(args_list=["https://example.com", "-l", "1"])
    c = crawler.Crawler.from_parser_args(args)
    urls = {"https://example.com", "https://www.iana.org/domains/example"}
    t: Set[str] = set()
    t = t.union(*c.run_crawl())
    if urls != t:
        raise AssertionError



#!/usr/bin/env python
import argparse
import http.client as httplib
import shutil
import sys
import textwrap
from typing import Iterable, List, Optional, Set

from wbsv import __version__
from wbsv.archiver import Archiver
from wbsv.crawler import Crawler


class HttpConnectionNotFountError(Exception):
    pass


class WBSVHelpFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
):
    pass


def check_connectivity(url: str = "www.google.com", timeout: int = 3) -> bool:
    conn = httplib.HTTPConnection(url, timeout=timeout)
    try:
        conn.request("HEAD", "/")
        conn.close()
        return True
    except Exception as e:
        print(e, file=sys.stderr)
        return False


def check_positive(v: str) -> int:
    if int(v) <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid natural number" % int(v))
    return int(v)


def parse_args(args_list: Optional[List[str]] = None) -> argparse.Namespace:
    """Parse arguments."""
    parser = argparse.ArgumentParser(
        prog="wbsv",
        formatter_class=(
            lambda prog: WBSVHelpFormatter(
                prog,
                **{
                    "width": shutil.get_terminal_size(fallback=(120, 50)).columns,
                    "max_help_position": 30,
                },
            )
        ),
        description=textwrap.dedent(
            """\
            CLI tool for save webpage on Wayback Machine forever.
            Save webpage and one 's all URI(s) on Wayback Machine."""
        ),
        epilog=textwrap.dedent(
            """\
            If you don't give the URL,
            interactive mode will be launched.
            (To quit interactive mode,
            type "end", "exit", "exit()",
            "break", "bye", ":q" or "finish".)"""
        ),
    )

    parser.add_argument(
        "url", metavar="url", nargs="*", type=str, help="Saving pages in order."
    )
    parser.add_argument(
        "-r",
        "--retry",
        type=check_positive,
        metavar="times",
        default=3,
        help="Set a retry limit on failed save.",
    )
    parser.add_argument(
        "-t",
        "--only_target",
        action="store_true",
        help="Save just same domain of target",
    )
    parser.add_argument(
        "-l",
        "--level",
        type=check_positive,
        metavar="level",
        default=1,
        help="Set maximum recursion depth",
    )
    parser.add_argument(
        "-O",
        "--own",
        action="store_true",
        help="Only URLs with the same domain as target",
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {__version__}"
    )
    if args_list:
        return parser.parse_args(args_list)
    else:
        return parser.parse_args()


def wbsv(
    urls: Iterable[str], own: bool, only_target: bool, level: int, retry: int
) -> None:
    past, now, fail = 0, 0, 0
    print(f"[+]Target: {urls}")
    c = Crawler.from_args(urls=urls, own=own, only_target=only_target, level=level)
    retrieved_links_empty_set: Set[str] = set()  # required for mypy type checking
    retrieved_links: Set[str] = retrieved_links_empty_set.union(*c.run_crawl())
    len_links: int = len(retrieved_links)
    len_len_links: int = len(str(len_links))
    print(f"[+]{len_links} URI(s) found.")
    a = Archiver(retry)
    for ind, link in enumerate(retrieved_links, 1):
        print(f"[{ind:0{len_len_links}d}/{len_links}]: Wait...", end="\r")
        archive = a.archive(link)
        if archive:
            archived_link, cached_flag = archive
            cache_or_now: str = "PAST" if cached_flag else "NOW"
            print(
                f"[{ind:0{len_len_links}d}/{len_links}]: <{cache_or_now}> {archived_link}"
            )
            inc = (1, 0) if cached_flag else (0, 1)
            past += inc[0]
            now += inc[1]
        else:
            print(f"[{ind:0{len_len_links}d}/{len_links}]: <FAIL> {link}")
            fail += 1

    print(f"[+]FIN!: {urls}")
    print(f"[+]ALL: {len_links}, NOW: {now}, PAST: {past}, FAIL: {fail}")


def wbsv_from_parser_args(args: argparse.Namespace) -> None:
    wbsv(args.url, args.own, args.only_target, args.level, args.retry)


def wbsv_repl(args: argparse.Namespace) -> None:
    finish_words = ("end", "exit", "exit()", "break", "bye", ":q", "finish")
    print("[[Input a target url (ex: https://google.com)]]")
    while True:
        link = input(">>> ").rstrip()
        if link in finish_words:
            print("[+]End.")
            break
        elif link != "":
            args.url = [link]
            try:
                wbsv_from_parser_args(args)
            except Exception as e:
                print(e, file=sys.stderr)


def _main() -> None:
    if not check_connectivity():
        raise HttpConnectionNotFountError
    args = parse_args()
    if len(sys.argv) <= 1:
        wbsv_repl(args)
    else:
        wbsv_from_parser_args(args)


def main() -> None:
    try:
        _main()
    except KeyboardInterrupt:
        exit(1)


if __name__ == "__main__":
    main()



import dataclasses
from abc import ABC, abstractmethod
from typing import Iterable, Set
from urllib.parse import ParseResult, urlparse


class UrlFilter(ABC):
    @abstractmethod
    def test_url(self, parsed_url: ParseResult) -> bool:
        pass


@dataclasses.dataclass
class SchemaFilter(UrlFilter):
    valid_schemas: Set[str] = dataclasses.field(
        default_factory=lambda: {"http", "https", "ftp", "file"}
    )

    def test_url(self, parsed_url: ParseResult) -> bool:
        return parsed_url.scheme in self.valid_schemas


@dataclasses.dataclass
class OwnDomainFilter(UrlFilter):
    own_domains: Set[str]

    @staticmethod
    def from_string_urls(urls: Iterable[str]) -> "OwnDomainFilter":
        return OwnDomainFilter({urlparse(url).netloc for url in urls})

    def test_url(self, parsed_url: ParseResult) -> bool:
        return parsed_url.netloc in self.own_domains


@dataclasses.dataclass
class CombinedFilter(UrlFilter):
    filters: Iterable[UrlFilter]

    def test_url(self, parsed_url: ParseResult) -> bool:
        return all(f.test_url(parsed_url) for f in self.filters)



from argparse import Namespace
from typing import Literal, Tuple, Union

import requests
import waybackpy  # type: ignore

from wbsv.utils import log_exception


class SavepagenowFailureError(Exception):
    pass


class Archiver:
    @staticmethod
    def from_parser_args(args: Namespace) -> "Archiver":
        return Archiver(args.retry)

    def __init__(self, retry: int):
        """Init."""
        self.retry: int = retry
        self.UA: str = (
            "Mozilla/5.0 (Windows NT 5.1; rv:40.0) " "Gecko/20100101 Firefox/40.0"
        )

    def archive(self, url: str) -> Union[Literal[False], Tuple[str, bool]]:
        """Archive link."""
        wp = waybackpy.WaybackMachineSaveAPI(url, self.UA, max_tries=self.retry)
        if self._try_savepagenow(wp):
            return wp.archive_url, wp.cached_save
        return False

    @staticmethod
    def _try_savepagenow(wp: waybackpy.WaybackMachineSaveAPI) -> bool:
        """Error handler for saving with savepagenow."""
        try:
            wp.save()
        except (
            waybackpy.exceptions.RedirectSaveError,
            waybackpy.exceptions.WaybackError,
            requests.exceptions.TooManyRedirects,
            AttributeError,
        ) as e:
            log_exception(e)
            return False
        else:
            return True
